/*
 * ***********************************************************************
 *
 * File 'tkCanvasPlusModes.c'.
 *
 * Tcl Interpreter for CanvasPlus widget.
 *
 * Written by Jorge Lima in Nov 23 14:44:03 WET 2004
 *
 */

#include <X11/Xutil.h>
#include <tcl.h>
#include <string.h>
#include <stdio.h>

#include "tkCanvasPlus.h"
#include "tkCanvasPlusModes.h"


// First element of the list is never used. Only there to be the head
// as it simplifies the implementation of insertion and deletion rotines
// ->next is initialized as NULL
/* Structure defining the mode */
typedef struct CP_ModeInfo {
  char *name;
  struct CP_RegisterInfo *info;
  struct CP_ModeOps *ops;
  struct CP_ModeInfo *next;
} CP_ModeInfo;

static CP_ModeInfo __ApplModes = {NULL, NULL, NULL, NULL};

/*
 * Register procedures
 */
static int RegisterProcs(Tcl_Interp *interp,
			 CP_RegisterInfo *records,
			 ClientData clientData)
{
  while ( records->procName != NULL ) {
    if ( records->function == NULL ) return TCL_ERROR;
    Tcl_CreateObjCommand(interp, records->procName,
			 records->function, clientData,
			 (Tcl_CmdDeleteProc *) NULL);
    records++;
  }
  return TCL_OK;
}


static void AppendModeElem(CP_ModeInfo *node)
{
  CP_ModeInfo *list = &__ApplModes;
  node->next = NULL;

  while (list->next != NULL) {
    list = list->next;
  }
  list->next = node;
}

/**************************************************************************
 * CPFindMode --
 *    Finds a mode with the given name
 * Returns --
 *    A pointer for the CPApplMode structure for that mode
 * Side efects --
 *    None
 **************************************************************************/
static CP_ModeInfo* CPFindMode(char *name)
{
  CP_ModeInfo *list = &__ApplModes;
  while (list->next != NULL) {
    list = list->next;
    if (strncmp(list->name, name, MODE_NAME_MAXLEN)==0)
      return list;
  }
  return NULL;
}

/**************************************************************************
 * CPDefineMode --
 *    Add a new mode to the mode list
 * Returns --
 *    A TCL return code
 * Side efects --
 **************************************************************************/
int CPDefineMode(char *name, CP_RegisterInfo *info, CP_ModeOps *ops)
{
  CP_ModeInfo *m = (CP_ModeInfo *)ckalloc(sizeof(CP_ModeInfo));
  if (m == NULL) return -1;
  m->name = name;
  m->info = info;
  m->ops = ops;
  AppendModeElem(m);
  return 0;
}

/**************************************************************************
 * CPListModes() --
 *    Build up a list of available modes
 * Returns --
 *    A TCL OBJ
 * Side efects --
 **************************************************************************/
Tcl_Obj *CPListModes()
{
  CP_ModeInfo *list = &__ApplModes;
  Tcl_Obj *obj = Tcl_NewObj();
  while (list->next != NULL) {
    list = list->next;
    Tcl_AppendToObj(obj, list->name, -1);
  }
  return obj;
}

/**************************************************************************
 * CPSetMode --
 *    Create a new interp and initialize it to a specific mode
 * Returns --
 *    A TCL return code
 * Side efects --
 *    In case of error, sets the interpreter result.
 *    In case of success a new interpreter is created, initialized and bound
 *    to the CanvasPlus pointer.
 **************************************************************************/
int CPSetMode(CanvasPlus *cPtr, Tcl_Obj *intName, Tcl_Obj *modeName)
{
  CP_ModeInfo *modeInfo;
  CP_Mode *mode;
  Tcl_Obj *resObj;

  // Check if interp already exists
  if (cPtr->mode != NULL) {
    resObj = Tcl_NewStringObj("canvas interp already initialized", -1);
    Tcl_SetObjResult(cPtr->interp, resObj);
    return TCL_ERROR;
  }
  // Check arguments
  if (intName == NULL || modeName == NULL) {
    resObj = Tcl_NewStringObj("INTERNAL ERROR: (Tcl_Obj *)arg==NULL,"
			      " where arg is intName or modeName", -1);
    Tcl_SetObjResult(cPtr->interp, resObj);
    return TCL_ERROR;
  }

  // Create and initialize mode structure
  modeInfo = CPFindMode(Tcl_GetString(modeName));

  if (modeInfo == NULL) {
    resObj = Tcl_NewStringObj("Can't find that mode",  -1);
    Tcl_SetObjResult(cPtr->interp, resObj);
    return TCL_ERROR;
  }

  // Alloc mode struct
  mode = (CP_Mode *)ckalloc((size_t)sizeof(CP_Mode));

  if (mode==NULL) {
    resObj = Tcl_NewStringObj("Can't alloc memory",  -1);
    Tcl_SetObjResult(cPtr->interp, resObj);
    return TCL_ERROR;
  }

  // initialize mode specific data
  mode->ops = modeInfo->ops;
  mode->clientData = mode->ops->allocProc(cPtr);

  if (mode->clientData == NULL) {
    resObj = Tcl_NewStringObj("Unable to initialize mode specific data.", -1);
    Tcl_SetObjResult(cPtr->interp, resObj);
    ckfree((void *)mode);
    return TCL_ERROR;
  }

  // Create slave interpreter
  mode->interp = Tcl_CreateSlave(cPtr->interp, Tcl_GetString(intName), 1);  

  if (mode->interp == NULL) {
    resObj = Tcl_NewStringObj("Unable to create slave interp.", -1);
    Tcl_SetObjResult(cPtr->interp, resObj);
    mode->ops->freeProc(mode->clientData);
    ckfree((void *)mode);
    return TCL_ERROR;
  }

  if( RegisterProcs(mode->interp, modeInfo->info, mode->clientData)==TCL_ERROR ) {
    resObj = Tcl_NewStringObj("Unable to register procs.", -1);
    Tcl_SetObjResult(cPtr->interp, resObj);
    Tcl_DeleteInterp(mode->interp);
    mode->ops->freeProc(mode->clientData);
    ckfree((void *)mode);
    return TCL_ERROR;
  }

  // Initialize the binding table (bindings are in the master interp)
  mode->bindingTable = Tk_CreateBindingTable(cPtr->interp);

  if (mode->bindingTable == NULL) {
    resObj = Tcl_NewStringObj("Unable to initialize binding table.", -1);
    Tcl_SetObjResult(cPtr->interp, resObj);
    Tcl_DeleteInterp(mode->interp);
    modeInfo->ops->freeProc(mode->clientData);
    ckfree((void *)mode);
    return TCL_ERROR;
  }
  // finish remaining fields
  mode->name = modeInfo->name;
  mode->tkwin = cPtr->tkwin;
  mode->firstItem = NULL;
  mode->lastItem = NULL;
  mode->nextId = 0;
  Tcl_InitHashTable(&mode->idTable, TCL_ONE_WORD_KEYS);
  CPInitTags();
  CPInitPenSet(mode);
  mode->ops->initProc(mode);
  cPtr->mode = mode;
  CPForceRedraw(cPtr);
  return TCL_OK;
}

/**************************************************************************
 * CPUnsetMode --
 *    Deletes the interp associated with this instance of CanvasPlus
 * Returns -
 *    None
 * Side efects --
 *    If privateInterp != NULL, the interpreter and all associated 
 *    data are deleted.
 **************************************************************************/
void CPUnsetMode(CanvasPlus *cPtr)
{
  if (cPtr->mode != NULL) {
    CP_Mode *mode = cPtr->mode;
    mode->nextId = 0;
    Tcl_DeleteHashTable(&mode->idTable);
    CPDeletePenSet(mode);
    CPReleaseTags();
    if (mode->clientData != NULL) {
      mode->ops->freeProc(mode->clientData);
    }
    if (mode->bindingTable != NULL) {
      Tk_DeleteBindingTable(mode->bindingTable);
    }
    if (mode->interp != NULL) {
      Tcl_DeleteInterp(mode->interp);
    }
    ckfree((void *)mode);
    cPtr->mode = NULL;
  }
}

/*
 * Names are truncated to 24 chars if longer 
 */
Tcl_Obj *CPGetUniqCmdName(Tcl_Interp *interp, char *base)
{
  char name[MODE_NAME_MAXLEN+1], basecopy[MODE_NAME_MAXLEN-1];
  int i = 0;
  Tcl_CmdInfo info;

  strncpy(basecopy, base, MODE_NAME_MAXLEN-2); 
  basecopy[MODE_NAME_MAXLEN-2]='\0';
  do {
    sprintf(name, "%s%d", basecopy, i); i++;
  } while (Tcl_GetCommandInfo(interp, name, &info));
  return Tcl_NewStringObj(name, -1);
}



